@@ -84,12 +84,6 @@
 #define DISCONNECT_TIMEOUT 1
 #define STREAM_TIMEOUT 20
 
-typedef enum {
-	AVDTP_SESSION_STATE_DISCONNECTED,
-	AVDTP_SESSION_STATE_CONNECTING,
-	AVDTP_SESSION_STATE_CONNECTED
-} avdtp_session_state_t;
-
 #if __BYTE_ORDER == __LITTLE_ENDIAN
 
 struct avdtp_common_header {
@@ -333,6 +327,12 @@ struct stream_callback {
 	unsigned int id;
 };
 
+struct avdtp_state_callback {
+	avdtp_session_state_cb cb;
+	void *user_data;
+	unsigned int id;
+};
+
 struct avdtp_stream {
 	GIOChannel *io;
 	uint16_t imtu;
@@ -360,7 +360,6 @@ struct avdtp {
 	struct avdtp_server *server;
 	bdaddr_t dst;
 
-	avdtp_session_state_t last_state;
 	avdtp_session_state_t state;
 
 	/* True if the session should be automatically disconnected */
@@ -401,6 +400,10 @@ struct avdtp {
 static GSList *servers = NULL;
 static GSList *sessions = NULL;
 
+static GSList *avdtp_callbacks = NULL;
+
+static gboolean auto_connect = TRUE;
+
 static int send_request(struct avdtp *session, gboolean priority,
 			struct avdtp_stream *stream, uint8_t signal_id,
 			void *buffer, size_t size);
@@ -611,7 +614,7 @@ static gboolean disconnect_timeout(gpointer user_data)
 	stream_setup = session->stream_setup;
 	session->stream_setup = FALSE;
 
-	dev = manager_get_device(&session->server->src, &session->dst);
+	dev = manager_get_device(&session->server->src, &session->dst, FALSE);
 
 	if (dev && dev->sink && stream_setup)
 		sink_setup_stream(dev->sink, session);
@@ -697,6 +700,29 @@ static struct avdtp_remote_sep *find_remote_sep(GSList *seps, uint8_t seid)
 	return NULL;
 }
 
+static void avdtp_set_state(struct avdtp *session,
+					avdtp_session_state_t new_state)
+{
+	GSList *l;
+	struct audio_device *dev;
+	bdaddr_t src, dst;
+	avdtp_session_state_t old_state = session->state;
+
+	session->state = new_state;
+
+	avdtp_get_peers(session, &src, &dst);
+	dev = manager_get_device(&src, &dst, FALSE);
+	if (dev == NULL) {
+		error("avdtp_set_state(): no matching audio device");
+		return;
+	}
+
+	for (l = avdtp_callbacks; l != NULL; l = l->next) {
+		struct avdtp_state_callback *cb = l->data;
+		cb->cb(dev, session, old_state, new_state, cb->user_data);
+	}
+}
+
 static void stream_free(struct avdtp_stream *stream)
 {
 	struct avdtp_remote_sep *rsep;
@@ -792,6 +818,37 @@ static void handle_transport_connect(struct avdtp *session, GIOChannel *io,
 					(GIOFunc) transport_cb, stream);
 }
 
+static int pending_req_cmp(gconstpointer a, gconstpointer b)
+{
+	const struct pending_req *req = a;
+	const struct avdtp_stream *stream = b;
+
+	if (req->stream == stream)
+		return 0;
+
+	return -1;
+}
+
+static void cleanup_queue(struct avdtp *session, struct avdtp_stream *stream)
+{
+	GSList *l;
+	struct pending_req *req;
+
+	while ((l = g_slist_find_custom(session->prio_queue, stream,
+							pending_req_cmp))) {
+		req = l->data;
+		pending_req_free(req);
+		session->prio_queue = g_slist_remove(session->prio_queue, req);
+	}
+
+	while ((l = g_slist_find_custom(session->req_queue, stream,
+							pending_req_cmp))) {
+		req = l->data;
+		pending_req_free(req);
+		session->req_queue = g_slist_remove(session->req_queue, req);
+	}
+}
+
 static void avdtp_sep_set_state(struct avdtp *session,
 				struct avdtp_local_sep *sep,
 				avdtp_state_t state)
@@ -799,6 +856,12 @@ static void avdtp_sep_set_state(struct avdtp *session,
 	struct avdtp_stream *stream = sep->stream;
 	avdtp_state_t old_state;
 	struct avdtp_error err, *err_ptr = NULL;
+	GSList *l;
+
+	if (!stream) {
+		error("Error changing sep state: stream not available");
+		return;
+	}
 
 	if (sep->state == state) {
 		avdtp_error_init(&err, AVDTP_ERROR_ERRNO, EIO);
@@ -814,13 +877,9 @@ static void avdtp_sep_set_state(struct avdtp *session,
 	old_state = sep->state;
 	sep->state = state;
 
-	if (stream) {
-		GSList *l;
-		for (l = stream->callbacks; l != NULL; l = g_slist_next(l)) {
-			struct stream_callback *cb = l->data;
-			cb->cb(stream, old_state, state, err_ptr,
-					cb->user_data);
-		}
+	for (l = stream->callbacks; l != NULL; l = g_slist_next(l)) {
+		struct stream_callback *cb = l->data;
+		cb->cb(stream, old_state, state, err_ptr, cb->user_data);
 	}
 
 	switch (state) {
@@ -848,6 +907,8 @@ static void avdtp_sep_set_state(struct avdtp *session,
 			handle_transport_connect(session, NULL, 0, 0);
 		if (session->req && session->req->stream == stream)
 			session->req->stream = NULL;
+		/* Remove pending commands for this stream from the queue */
+		cleanup_queue(session, stream);
 		stream_free(stream);
 		if (session->ref == 1 && !session->streams)
 			set_disconnect_timer(session);
@@ -886,16 +947,8 @@ static void release_stream(struct avdtp_stream *stream, struct avdtp *session)
 
 static void connection_lost(struct avdtp *session, int err)
 {
-	struct audio_device *dev;
 	char address[18];
 
-	dev = manager_find_device(&session->dst, AUDIO_CONTROL_INTERFACE,
-					FALSE);
-	if (dev && dev->control) {
-		device_remove_control_timer(dev);
-		avrcp_disconnect(dev);
-	}
-
 	ba2str(&session->dst, address);
 	debug("Disconnected from %s", address);
 
@@ -917,7 +970,7 @@ static void connection_lost(struct avdtp *session, int err)
 		session->io = NULL;
 	}
 
-	session->state = AVDTP_SESSION_STATE_DISCONNECTED;
+	avdtp_set_state(session, AVDTP_SESSION_STATE_DISCONNECTED);
 
 	if (session->io_id) {
 		g_source_remove(session->io_id);
@@ -1105,8 +1158,8 @@ static gboolean avdtp_getcap_cmd(struct avdtp *session, uint8_t transaction,
 	uint8_t err, buf[1024], *ptr = buf;
 
 	if (size < sizeof(struct seid_req)) {
-		error("Too short getcap request");
-		return FALSE;
+		err = AVDTP_BAD_LENGTH;
+		goto failed;
 	}
 
 	sep = find_local_sep_by_seid(session->server, req->acp_seid);
@@ -1151,6 +1204,7 @@ static gboolean avdtp_setconf_cmd(struct avdtp *session, uint8_t transaction,
 	uint8_t err, category = 0x00;
 	struct audio_device *dev;
 	bdaddr_t src, dst;
+	GSList *l;
 
 	if (size < sizeof(struct setconf_req)) {
 		error("Too short getcap request");
@@ -1169,7 +1223,7 @@ static gboolean avdtp_setconf_cmd(struct avdtp *session, uint8_t transaction,
 	}
 
 	avdtp_get_peers(session, &src, &dst);
-	dev = manager_get_device(&src, &dst);
+	dev = manager_get_device(&src, &dst, FALSE);
 	if (!dev) {
 		error("Unable to get a audio device object");
 		goto failed;
@@ -1198,6 +1252,16 @@ static gboolean avdtp_setconf_cmd(struct avdtp *session, uint8_t transaction,
 					size - sizeof(struct setconf_req),
 					&stream->codec);
 
+	/* Verify that the Media Transport capability's length = 0. Reject otherwise */
+	for (l = stream->caps; l != NULL; l = g_slist_next(l)) {
+		struct avdtp_service_capability *cap = l->data;
+
+		if (cap->category == AVDTP_MEDIA_TRANSPORT && cap->length != 0) {
+			err = AVDTP_BAD_MEDIA_TRANSPORT_FORMAT;
+			goto failed;
+		}
+	}
+
 	if (sep->ind && sep->ind->set_configuration) {
 		if (!sep->ind->set_configuration(session, sep, stream,
 							stream->caps, &err,
@@ -1231,7 +1295,47 @@ failed:
 static gboolean avdtp_getconf_cmd(struct avdtp *session, uint8_t transaction,
 					struct seid_req *req, int size)
 {
-	return avdtp_unknown_cmd(session, transaction, (void *) req, size);
+	GSList *l;
+	struct avdtp_local_sep *sep = NULL;
+	int rsp_size;
+	uint8_t err;
+	uint8_t buf[1024];
+	uint8_t *ptr = buf;
+
+	if (size < (int) sizeof(struct seid_req)) {
+		error("Too short getconf request");
+		return FALSE;
+	}
+
+	memset(buf, 0, sizeof(buf));
+
+	sep = find_local_sep_by_seid(session->server, req->acp_seid);
+	if (!sep) {
+		err = AVDTP_BAD_ACP_SEID;
+		goto failed;
+	}
+	if (!sep->stream || !sep->stream->caps) {
+		err = AVDTP_UNSUPPORTED_CONFIGURATION;
+		goto failed;
+	}
+
+	for (l = sep->stream->caps, rsp_size = 0; l != NULL; l = g_slist_next(l)) {
+		struct avdtp_service_capability *cap = l->data;
+
+		if (rsp_size + cap->length + 2 > (int) sizeof(buf))
+			break;
+
+		memcpy(ptr, cap, cap->length + 2);
+		rsp_size += cap->length + 2;
+		ptr += cap->length + 2;
+	}
+
+	return avdtp_send(session, transaction, AVDTP_MSG_TYPE_ACCEPT,
+				AVDTP_GET_CONFIGURATION, buf, rsp_size);
+
+failed:
+	return avdtp_send(session, transaction, AVDTP_MSG_TYPE_REJECT,
+				AVDTP_GET_CONFIGURATION, &err, sizeof(err));
 }
 
 static gboolean avdtp_reconf_cmd(struct avdtp *session, uint8_t transaction,
@@ -1807,6 +1911,8 @@ static struct avdtp *avdtp_get_internal(const bdaddr_t *src, const bdaddr_t *dst
 	session->server = find_server(servers, src);
 	bacpy(&session->dst, dst);
 	session->ref = 1;
+	/* We don't use avdtp_set_state() here since this isn't a state change
+	 * but just setting of the initial state */
 	session->state = AVDTP_SESSION_STATE_DISCONNECTED;
 	session->auto_dc = TRUE;
 
@@ -1843,10 +1949,8 @@ static void avdtp_connect_cb(GIOChannel *chan, GError *err, gpointer user_data)
 		goto failed;
 	}
 
-	if (session->state == AVDTP_SESSION_STATE_DISCONNECTED) {
+	if (!session->io)
 		session->io = g_io_channel_ref(chan);
-		session->state = AVDTP_SESSION_STATE_CONNECTING;
-	}
 
 	bt_io_get(chan, BT_IO_L2CAP, &gerr,
 			BT_IO_OPT_OMTU, &session->omtu,
@@ -1864,30 +1968,32 @@ static void avdtp_connect_cb(GIOChannel *chan, GError *err, gpointer user_data)
 			address);
 
 	if (session->state == AVDTP_SESSION_STATE_CONNECTING) {
-		struct audio_device *dev;
-
 		debug("AVDTP imtu=%u, omtu=%u", session->imtu, session->omtu);
 
 		session->buf = g_malloc0(session->imtu);
-		session->state = AVDTP_SESSION_STATE_CONNECTED;
-		session->io_id = g_io_add_watch(chan,
+		avdtp_set_state(session, AVDTP_SESSION_STATE_CONNECTED);
+
+		if (session->io_id)
+			g_source_remove(session->io_id);
+
+		/* This watch should be low priority since otherwise the
+		 * connect callback might be dispatched before the session
+		 * callback if the kernel wakes us up at the same time for
+		 * them. This could happen if a headset is very quick in
+		 * sending the Start command after connecting the stream
+		 * transport channel.
+		 */
+		session->io_id = g_io_add_watch_full(chan,
+						G_PRIORITY_LOW,
 						G_IO_IN | G_IO_ERR | G_IO_HUP
 						| G_IO_NVAL,
-						(GIOFunc) session_cb, session);
+						(GIOFunc) session_cb, session,
+						NULL);
 
 		if (session->stream_setup) {
 			set_disconnect_timer(session);
 			avdtp_set_auto_disconnect(session, FALSE);
 		}
-
-		dev = manager_find_device(&session->dst,
-					AUDIO_CONTROL_INTERFACE, FALSE);
-		if (dev && dev->control) {
-			if (session->stream_setup)
-				device_set_control_timer(dev);
-			else
-				avrcp_connect(dev);
-		}
 	} else if (session->pending_open)
 		handle_transport_connect(session, chan, session->imtu,
 								session->omtu);
@@ -1932,15 +2038,13 @@ static void auth_cb(DBusError *derr, void *user_data)
 		return;
 	}
 
-	/* Here we set the disconnect timer so we don't stay in IDLE state
-	 * indefinitely but set auto_dc to FALSE so that when a stream is
-	 * finally opened it doesn't get closed due to a timeout */
+	/* This is so that avdtp_connect_cb will know to do the right thing
+	 * with respect to the disconnect timer */
 	session->stream_setup = TRUE;
 }
 
 static void avdtp_confirm_cb(GIOChannel *chan, gpointer data)
 {
-	int sk;
 	struct avdtp *session;
 	struct audio_device *dev;
 	char address[18];
@@ -1962,8 +2066,8 @@ static void avdtp_confirm_cb(GIOChannel *chan, gpointer data)
 	debug("AVDTP: incoming connect from %s", address);
 
 	session = avdtp_get_internal(&src, &dst);
-
-	sk = g_io_channel_unix_get_fd(chan);
+	if (!session)
+		goto drop;
 
 	if (session->pending_open && session->pending_open->open_acp) {
 		if (!bt_io_accept(chan, avdtp_connect_cb, session, NULL, NULL))
@@ -1976,14 +2080,14 @@ static void avdtp_confirm_cb(GIOChannel *chan, gpointer data)
 		goto drop;
 	}
 
-	dev = manager_get_device(&src, &dst);
+	dev = manager_get_device(&src, &dst, TRUE);
 	if (!dev) {
 		error("Unable to get audio device object for %s", address);
 		goto drop;
 	}
 
 	session->io = g_io_channel_ref(chan);
-	session->state = AVDTP_SESSION_STATE_CONNECTING;
+	avdtp_set_state(session, AVDTP_SESSION_STATE_CONNECTING);
 
 	session->io_id = g_io_add_watch(chan, G_IO_ERR | G_IO_HUP | G_IO_NVAL,
 					(GIOFunc) session_cb, session);
@@ -1994,6 +2098,8 @@ static void avdtp_confirm_cb(GIOChannel *chan, gpointer data)
 		goto drop;
 	}
 
+	dev->auto_connect = auto_connect;
+
 	return;
 
 drop:
@@ -2131,6 +2237,7 @@ static int send_req(struct avdtp *session, gboolean priority,
 		err = l2cap_connect(session);
 		if (err < 0)
 			goto failed;
+		avdtp_set_state(session, AVDTP_SESSION_STATE_CONNECTING);
 	}
 
 	if (session->state < AVDTP_SESSION_STATE_CONNECTED ||
@@ -2448,7 +2555,7 @@ static gboolean stream_rej_to_err(struct stream_rej *rej, unsigned int size,
 					struct avdtp_error *err,
 					uint8_t *acp_seid)
 {
-	if (size < sizeof(struct conf_rej)) {
+	if (size < sizeof(struct stream_rej)) {
 		error("Too small packet for stream_rej");
 		return FALSE;
 	}
@@ -2567,6 +2674,21 @@ gboolean avdtp_is_connected(const bdaddr_t *src, const bdaddr_t *dst)
 	return FALSE;
 }
 
+struct avdtp_service_capability *avdtp_stream_get_codec(
+						struct avdtp_stream *stream)
+{
+	GSList *l;
+
+	for (l = stream->caps; l; l = l->next) {
+		struct avdtp_service_capability *cap = l->data;
+
+		if (cap->category == AVDTP_MEDIA_CODEC)
+			return cap;
+	}
+
+	return NULL;
+}
+
 gboolean avdtp_stream_has_capability(struct avdtp_stream *stream,
 				struct avdtp_service_capability *cap)
 {
@@ -2575,11 +2697,13 @@ gboolean avdtp_stream_has_capability(struct avdtp_stream *stream,
 
 	for (l = stream->caps; l; l = g_slist_next(l)) {
 		stream_cap = l->data;
-		if (stream_cap->category == cap->category &&
-			stream_cap->length == cap->length) {
-			if (!memcmp(stream_cap->data, cap->data, cap->length))
-				return TRUE;
-		}
+
+		if (stream_cap->category != cap->category ||
+			stream_cap->length != cap->length)
+			continue;
+
+		if (memcmp(stream_cap->data, cap->data, cap->length) == 0)
+			return TRUE;
 	}
 
 	return FALSE;
@@ -2646,11 +2770,41 @@ static int process_queue(struct avdtp *session)
 	return send_req(session, FALSE, req);
 }
 
+struct avdtp_remote_sep *avdtp_get_remote_sep(struct avdtp *session,
+						uint8_t seid)
+{
+	GSList *l;
+
+	for (l = session->seps; l; l = l->next) {
+		struct avdtp_remote_sep *sep = l->data;
+
+		if (sep->seid == seid)
+			return sep;
+	}
+
+	return NULL;
+}
+
+uint8_t avdtp_get_seid(struct avdtp_remote_sep *sep)
+{
+	return sep->seid;
+}
+
+uint8_t avdtp_get_type(struct avdtp_remote_sep *sep)
+{
+	return sep->type;
+}
+
 struct avdtp_service_capability *avdtp_get_codec(struct avdtp_remote_sep *sep)
 {
 	return sep->codec;
 }
 
+struct avdtp_stream *avdtp_get_stream(struct avdtp_remote_sep *sep)
+{
+	return sep->stream;
+}
+
 struct avdtp_service_capability *avdtp_service_cap_new(uint8_t category,
 							void *data, int length)
 {
@@ -2667,6 +2821,15 @@ struct avdtp_service_capability *avdtp_service_cap_new(uint8_t category,
 	return cap;
 }
 
+static gboolean process_discover(gpointer data)
+{
+	struct avdtp *session = data;
+
+	finalize_discovery(session, 0);
+
+	return FALSE;
+}
+
 int avdtp_discover(struct avdtp *session, avdtp_discover_cb_t cb,
 			void *user_data)
 {
@@ -2676,7 +2839,9 @@ int avdtp_discover(struct avdtp *session, avdtp_discover_cb_t cb,
 		return -EBUSY;
 
 	if (session->seps) {
-		cb(session, session->seps, NULL, user_data);
+		session->discov_cb = cb;
+		session->user_data = user_data;
+		g_idle_add(process_discover, session);
 		return 0;
 	}
 
@@ -3150,9 +3315,16 @@ int avdtp_init(const bdaddr_t *src, GKeyFile *config)
 							"Master", &err);
 		if (err) {
 			debug("audio.conf: %s", err->message);
-			g_error_free(err);
+			g_clear_error(&err);
 		} else
 			master = tmp;
+
+		tmp = g_key_file_get_boolean(config, "General", "AutoConnect",
+									&err);
+		if (err)
+			g_clear_error(&err);
+		else
+			auto_connect = tmp;
 	}
 
 	server = g_new0(struct avdtp_server, 1);
@@ -3196,3 +3368,39 @@ void avdtp_set_auto_disconnect(struct avdtp *session, gboolean auto_dc)
 {
 	session->auto_dc = auto_dc;
 }
+
+gboolean avdtp_stream_setup_active(struct avdtp *session)
+{
+	return session->stream_setup;
+}
+
+unsigned int avdtp_add_state_cb(avdtp_session_state_cb cb, void *user_data)
+{
+	struct avdtp_state_callback *state_cb;
+	static unsigned int id = 0;
+
+	state_cb = g_new(struct avdtp_state_callback, 1);
+	state_cb->cb = cb;
+	state_cb->user_data = user_data;
+	state_cb->id = ++id;
+
+	avdtp_callbacks = g_slist_append(avdtp_callbacks, state_cb);;
+
+	return state_cb->id;
+}
+
+gboolean avdtp_remove_state_cb(unsigned int id)
+{
+	GSList *l;
+
+	for (l = avdtp_callbacks; l != NULL; l = l->next) {
+		struct avdtp_state_callback *cb = l->data;
+		if (cb && cb->id == id) {
+			avdtp_callbacks = g_slist_remove(avdtp_callbacks, cb);
+			g_free(cb);
+			return TRUE;
+		}
+	}
+
+	return FALSE;
+}
