@@ -52,6 +52,7 @@
 #include <gdbus.h>
 
 #include "../src/dbus-common.h"
+#include "../src/adapter.h"
 
 #include "logging.h"
 #include "textfile.h"
@@ -60,6 +61,7 @@
 #include "storage.h"
 #include "sdpd.h"
 #include "glib-helper.h"
+#include "btio.h"
 #include "proxy.h"
 
 #define SERIAL_PORT_NAME	"spp"
@@ -98,8 +100,8 @@ struct serial_proxy {
 	uint8_t		channel;	/* RFCOMM channel */
 	uint32_t	record_id;	/* Service record id */
 	GIOChannel	*io;		/* Server listen */
-	guint		rfcomm_watch;	/* RFCOMM watch: Remote */
-	guint		local_watch;	/* Local watch: TTY or Unix socket */
+	GIOChannel	*rfcomm;	/* Remote RFCOMM channel*/
+	GIOChannel	*local;		/* Local channel: TTY or Unix socket */
 	struct serial_adapter *adapter;	/* Adapter pointer */
 };
 
@@ -108,14 +110,16 @@ static int sk_counter = 0;
 
 static void disable_proxy(struct serial_proxy *prx)
 {
-	if (prx->rfcomm_watch) {
-		g_source_remove(prx->rfcomm_watch);
-		prx->rfcomm_watch = 0;
+	if (prx->rfcomm) {
+		g_io_channel_shutdown(prx->rfcomm, TRUE, NULL);
+		g_io_channel_unref(prx->rfcomm);
+		prx->rfcomm = NULL;
 	}
 
-	if (prx->local_watch) {
-		g_source_remove(prx->local_watch);
-		prx->local_watch = 0;
+	if (prx->local) {
+		g_io_channel_shutdown(prx->local, TRUE, NULL);
+		g_io_channel_unref(prx->local);
+		prx->local = NULL;
 	}
 
 	remove_record_from_server(prx->record_id);
@@ -246,13 +250,16 @@ static GIOError channel_write(GIOChannel *chan, char *buf, size_t size)
 static gboolean forward_data(GIOChannel *chan, GIOCondition cond, gpointer data)
 {
 	char buf[BUF_SIZE];
-	GIOChannel *dest = data;
+	struct serial_proxy *prx = data;
+	GIOChannel *dest;
 	GIOError err;
 	size_t rbytes;
 
 	if (cond & G_IO_NVAL)
 		return FALSE;
 
+	dest = (chan == prx->rfcomm) ? prx->local : prx->rfcomm;
+
 	if (cond & (G_IO_HUP | G_IO_ERR)) {
 		/* Try forward remaining data */
 		do {
@@ -264,7 +271,14 @@ static gboolean forward_data(GIOChannel *chan, GIOCondition cond, gpointer data)
 			err = channel_write(dest, buf, rbytes);
 		} while (err == G_IO_ERROR_NONE);
 
-		g_io_channel_close(dest);
+		g_io_channel_shutdown(prx->local, TRUE, NULL);
+		g_io_channel_unref(prx->local);
+		prx->local = NULL;
+
+		g_io_channel_shutdown(prx->rfcomm, TRUE, NULL);
+		g_io_channel_unref(prx->rfcomm);
+		prx->rfcomm = NULL;
+
 		return FALSE;
 	}
 
@@ -382,20 +396,17 @@ static inline int tty_open(const char *tty, struct termios *ti)
 	return sk;
 }
 
-static void connect_event_cb(GIOChannel *chan, int err, const bdaddr_t *src,
-				const bdaddr_t *dst, gpointer data)
+static void connect_event_cb(GIOChannel *chan, GError *conn_err, gpointer data)
 {
 	struct serial_proxy *prx = data;
-	GIOChannel *io;
 	int sk;
 
-	if (err < 0) {
-		error("accept: %s (%d)", strerror(-err), -err);
-		return;
+	if (conn_err) {
+		error("%s", conn_err->message);
+		goto drop;
 	}
 
-	bacpy(&prx->dst, dst);
-
+	/* Connect local */
 	switch (prx->type) {
 	case UNIX_SOCKET_PROXY:
 		sk = unix_socket_connect(prx->address);
@@ -410,27 +421,93 @@ static void connect_event_cb(GIOChannel *chan, int err, const bdaddr_t *src,
 		sk = -1;
 	}
 
-	if (sk < 0) {
-		g_io_channel_unref(chan);
-		return;
-	}
+	if (sk < 0)
+		goto drop;
 
-	g_io_channel_set_close_on_unref(chan, TRUE);
-	io = g_io_channel_unix_new(sk);
-	g_io_channel_set_close_on_unref(io, TRUE);
+	prx->local = g_io_channel_unix_new(sk);
 
-	prx->rfcomm_watch = g_io_add_watch(chan,
-				G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
-				forward_data, io);
+	g_io_add_watch(prx->rfcomm,
+			G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
+			forward_data, prx);
 
-	prx->local_watch = g_io_add_watch(io,
-				G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
-				forward_data, chan);
-
-	g_io_channel_unref(chan);
-	g_io_channel_unref(io);
+	g_io_add_watch(prx->local,
+			G_IO_IN | G_IO_HUP | G_IO_ERR | G_IO_NVAL,
+			forward_data, prx);
 
 	return;
+
+drop:
+	g_io_channel_shutdown(prx->rfcomm, TRUE, NULL);
+	g_io_channel_unref(prx->rfcomm);
+	prx->rfcomm = NULL;
+}
+
+static void auth_cb(DBusError *derr, void *user_data)
+{
+	struct serial_proxy *prx = user_data;
+	GError *err = NULL;
+
+	if (derr) {
+		error("Access denied: %s", derr->message);
+		goto reject;
+	}
+
+	if (!bt_io_accept(prx->rfcomm, connect_event_cb, prx, NULL,
+							&err)) {
+		error("bt_io_accept: %s", err->message);
+		g_error_free(err);
+		goto reject;
+	}
+
+	return;
+
+reject:
+	g_io_channel_shutdown(prx->rfcomm, TRUE, NULL);
+	g_io_channel_unref(prx->rfcomm);
+	prx->rfcomm = NULL;
+}
+
+static void confirm_event_cb(GIOChannel *chan, gpointer user_data)
+{
+	struct serial_proxy *prx = user_data;
+	int perr;
+	char *address[18];
+	GError *err = NULL;
+
+	bt_io_get(chan, BT_IO_RFCOMM, &err,
+			BT_IO_OPT_DEST_BDADDR, &prx->dst,
+			BT_IO_OPT_DEST, address,
+			BT_IO_OPT_INVALID);
+	if (err) {
+		error("%s", err->message);
+		g_error_free(err);
+		goto drop;
+	}
+
+	if (prx->rfcomm) {
+		error("Refusing connect from %s: Proxy already in use",
+				address);
+		goto drop;
+	}
+
+	debug("Serial Proxy: incoming connect from %s", address);
+
+	prx->rfcomm = g_io_channel_ref(chan);
+
+	perr = btd_request_authorization(&prx->adapter->src, &prx->dst,
+					prx->uuid128, auth_cb, prx);
+	if (perr < 0) {
+		error("Refusing connect from %s: %s (%d)", address,
+				strerror(-perr), -perr);
+		g_io_channel_unref(prx->rfcomm);
+		prx->rfcomm = NULL;
+		goto drop;
+	}
+
+	return;
+
+drop:
+	g_io_channel_shutdown(chan, TRUE, NULL);
 }
 
 static DBusMessage *proxy_enable(DBusConnection *conn,
@@ -439,19 +516,31 @@ static DBusMessage *proxy_enable(DBusConnection *conn,
 	struct serial_proxy *prx = data;
 	struct serial_adapter *adapter = prx->adapter;
 	sdp_record_t *record;
+	GError *err = NULL;
+	DBusMessage *reply;
 
 	if (prx->io)
 		return failed(msg, "Already enabled");
 
 	/* Listen */
-	prx->io = bt_rfcomm_listen_allocate(&adapter->src, &prx->channel, 0,
-				connect_event_cb, prx);
-	if (!prx->io) {
-		const char *strerr = strerror(errno);
-		error("RFCOMM listen socket failed: %s(%d)", strerr, errno);
-		return failed(msg, strerr);
+	prx->io = bt_io_listen(BT_IO_RFCOMM, NULL, confirm_event_cb, prx,
+				NULL, &err,
+				BT_IO_OPT_SOURCE_BDADDR, &adapter->src,
+				BT_IO_OPT_INVALID);
+	if (!prx->io)
+		goto failed;
+
+	bt_io_get(prx->io, BT_IO_RFCOMM, &err,
+			BT_IO_OPT_CHANNEL, &prx->channel,
+			BT_IO_OPT_INVALID);
+	if (err) {
+		g_io_channel_unref(prx->io);
+		prx->io = NULL;
+		goto failed;
 	}
 
+	debug("Allocated channel %d", prx->channel);
+
 	g_io_channel_set_close_on_unref(prx->io, TRUE);
 
 	record = proxy_record_new(prx->uuid128, prx->channel);
@@ -469,6 +558,12 @@ static DBusMessage *proxy_enable(DBusConnection *conn,
 	prx->record_id = record->handle;
 
 	return dbus_message_new_method_return(msg);
+
+failed:
+	error("%s", err->message);
+	reply = failed(msg, err->message);
+	g_error_free(err);
+	return reply;
 }
 
 static DBusMessage *proxy_disable(DBusConnection *conn,
@@ -515,7 +610,7 @@ static DBusMessage *proxy_get_info(DBusConnection *conn,
 	boolean = (prx->io ? TRUE : FALSE);
 	dict_append_entry(&dict, "enabled", DBUS_TYPE_BOOLEAN, &boolean);
 
-	boolean = (prx->rfcomm_watch ? TRUE : FALSE);
+	boolean = (prx->rfcomm ? TRUE : FALSE);
 	dict_append_entry(&dict, "connected", DBUS_TYPE_BOOLEAN, &boolean);
 
 	/* If connected: append the remote address */
@@ -637,7 +732,7 @@ static DBusMessage *proxy_set_serial_params(DBusConnection *conn,
 	speed_t speed = B0;	/* In/Out speed */
 
 	/* Don't allow change TTY settings if it is open */
-	if (prx->local_watch)
+	if (prx->local)
 		return failed(msg, "Not allowed");
 
 	if (!dbus_message_get_args(msg, NULL,
@@ -685,7 +780,7 @@ static void proxy_path_unregister(gpointer data)
 	struct serial_proxy *prx = data;
 	int sk;
 
-	info("Unregistered proxy: %s", prx->address);
+	debug("Unregistered proxy: %s", prx->address);
 
 	if (prx->type != TTY_PROXY)
 		goto done;
@@ -723,7 +818,7 @@ static int register_proxy_object(struct serial_proxy *prx, char *outpath, size_t
 	if (outpath)
 		strncpy(outpath, path, size);
 
-	info("Registered proxy: %s", path);
+	debug("Registered proxy: %s", path);
 
 	return 0;
 }
@@ -870,65 +965,66 @@ static DBusMessage *create_proxy(DBusConnection *conn,
 {
 	struct serial_adapter *adapter = data;
 	char path[MAX_PATH_LENGTH + 1];
-	const char *uuid128, *address, *ppath = path;
-	DBusMessage *reply;
+	const char *pattern, *address, *ppath = path;
+	char *uuid_str;
 	proxy_type_t type;
 	uuid_t uuid;
 	int ret;
 
 	if (!dbus_message_get_args(msg, NULL,
-				DBUS_TYPE_STRING, &uuid128,
+				DBUS_TYPE_STRING, &pattern,
 				DBUS_TYPE_STRING, &address,
 				DBUS_TYPE_INVALID))
 		return NULL;
 
-	if (bt_string2uuid(&uuid, uuid128) < 0)
+	uuid_str = bt_name2string(pattern);
+	if (!uuid_str)
 		return invalid_arguments(msg, "Invalid UUID");
 
+	bt_string2uuid(&uuid, uuid_str);
+
 	type = addr2type(address);
-	if (type == UNKNOWN_PROXY_TYPE)
+	if (type == UNKNOWN_PROXY_TYPE) {
+		g_free(uuid_str);
 		return invalid_arguments(msg, "Invalid address");
+	}
 
 	/* Only one proxy per address(TTY or unix socket) is allowed */
-	if (g_slist_find_custom(adapter->proxies, address, proxy_addrcmp))
+	if (g_slist_find_custom(adapter->proxies, address, proxy_addrcmp)) {
+		g_free(uuid_str);
 		return g_dbus_create_error(msg, ERROR_INTERFACE ".AlreadyExist",
 						"Proxy already exists");
-
-	reply = dbus_message_new_method_return(msg);
-	if (!reply)
-		return NULL;
+	}
 
 	switch (type) {
 	case UNIX_SOCKET_PROXY:
-		ret = proxy_socket_register(adapter, uuid128, address,
+		ret = proxy_socket_register(adapter, uuid_str, address,
 						path, sizeof(path), TRUE);
 		break;
 	case TTY_PROXY:
-		ret = proxy_tty_register(adapter, uuid128, address,
+		ret = proxy_tty_register(adapter, uuid_str, address,
 				NULL, path, sizeof(path), TRUE);
 		break;
 	case TCP_SOCKET_PROXY:
-		ret = proxy_tcp_register(adapter, uuid128, address,
+		ret = proxy_tcp_register(adapter, uuid_str, address,
 					path, sizeof(path), TRUE);
 		break;
 	default:
 		ret = -1;
 	}
-	if (ret < 0) {
-		dbus_message_unref(reply);
+
+	g_free(uuid_str);
+
+	if (ret < 0)
 		return failed(msg, "Create object path failed");
-	}
 
 	g_dbus_emit_signal(adapter->conn, adapter->path,
 			SERIAL_MANAGER_INTERFACE, "ProxyCreated",
 			DBUS_TYPE_STRING, &ppath,
 			DBUS_TYPE_INVALID);
 
-	dbus_message_append_args(reply,
-			DBUS_TYPE_STRING, &ppath,
-			DBUS_TYPE_INVALID);
-
-	return reply;
+	return g_dbus_create_reply(msg, DBUS_TYPE_STRING, &ppath,
+					DBUS_TYPE_INVALID);
 }
 
 static DBusMessage *list_proxies(DBusConnection *conn,
@@ -1033,7 +1129,8 @@ static void parse_proxy(char *key, char *value, void *data)
 	char uuid128[MAX_LEN_UUID_STR], tmp[3];
 	char *pvalue;
 	proxy_type_t type;
-	int ch, opts, pos;
+	unsigned int pos;
+	int ch, opts;
 	struct termios ti;
 	uint8_t *pti;
 
@@ -1132,7 +1229,7 @@ int proxy_register(DBusConnection *conn, const char *path, bdaddr_t *src)
 
 	adapters = g_slist_append(adapters, adapter);
 
-	info("Registered interface %s on path %s",
+	debug("Registered interface %s on path %s",
 		SERIAL_MANAGER_INTERFACE, path);
 
 	return 0;
